using System;
using System.Collections.Generic;
using System.Text;

namespace PERQDisk
{
    // Represents a single or double-sided 8" floppy disk.
    public class FloppyDisk : PhysicalDisk
    {
        public FloppyDisk(bool isDoubleSided) : base()
        {
            _diskType = isDoubleSided ? DiskGeometry.FloppyDS : DiskGeometry.FloppySS;
            _header = new BlockHeader[_diskType.Cylinders, _diskType.Tracks, 6];

            CreateSectors();
        }

        public override void Load(System.IO.FileStream fs)
        {
            base.Load(fs);

            // Fill in header data for each logical block from first sector of every cylinder...


        }

        /// <summary>
        /// Returns a hacked-up sector containing 4 64-word sectors (one logical block)
        /// which is the smallest unit of disk access used by the POS filesystem.
        /// </summary>
        /// <param name="cylinder"></param>
        /// <param name="track"></param>
        /// <param name="sector"></param>
        /// <returns></returns>
        public override Sector GetSector(uint cylinder, uint track, uint sector)
        {
            if (cylinder < 0 || cylinder >= _diskType.Cylinders)
            {
                throw new ArgumentOutOfRangeException("cylinder");
            }

            return CreateLogicalBlock(cylinder, track, sector);
        }

        /// <summary>
        /// Returns a hacked-up sector containing 4 64-word sectors (one logical block)
        /// which is the smallest unit of disk access used by the POS filesystem.
        /// </summary>
        /// <param name="cylinder"></param>
        /// <param name="track"></param>
        /// <param name="sector"></param>
        /// <returns></returns>
        public override Sector GetSector(Block block)
        {
            // Get physical block if this is a logical one.
            if (block.IsLogical)
            {
                block = LogicalBlockToPhysicalBlock(block);
            }

            return CreateLogicalBlock(block.Cylinder, block.Track, block.Sector);
        }

        /// <summary>
        /// Converts a logical disk address into a Logical block expressed in Cyl,Track,Sector format.
        /// 
        /// This routine converts a LDA (logical disk address) to an LBN; this can then be directly converted
        /// to a Physical block address (LBN's cyl0,track0,sector0 is physical cyl5,track0,sector2)
        /// Summarized from the POS documentation, Page 1-7:       
        /// 
        /// Low word:
        ///     - Bits 0 through 7 are reserved for future use
        ///     - Bits 8 through 15 form the low 8 bits of the LBN
        /// 
        /// High word:
        ///     - Bits 0 through 10 form the high eleven bits of the LBN
        ///     - Bits 11 through 13 specify the volume (000 = hard disk, 100 = floppy)
        ///     - Bits 14 and 15 indicate whether the address is on disk or virtualized.
        /// 
        /// Currently we assume a hard disk image and so only the LBN-specific data is taken into account here.
        ///         
        /// 
        /// </summary>
        /// <param name="logicalAddress"></param>
        /// <returns></returns>
        public override Block LDAToLogicalBlock(uint lda)
        {
            uint lba = ((lda & 0x0000ff00) >> 8) | ((lda & 0x07ff0000) >> 8);

            //Console.WriteLine("lda: {0:x}", lda ); //(lda & 0x38000000) >> 27);
            //Console.WriteLine("lba: {0}", lba);

            //Now we have the actual logical block number. 
            uint logicalSectors = ((Sectors - 2) / 4);
            
            if (lba > (Cylinders - 2) * logicalSectors * Tracks)
            {
                Console.WriteLine((String.Format("Warning: Invalid LBA {0}", lba)));
            }

            Block pb = new Block();
            pb.Track = lba >= ((Cylinders - 5) * logicalSectors) ? (uint)1 : (uint)0;

            if( pb.Track == 0 )
            {
                pb.Cylinder = (uint)(lba / logicalSectors) % (Cylinders - 5);
            }
            else
            {
                pb.Cylinder = (uint)((lba / logicalSectors) % (Cylinders));
            }

            pb.Sector = (uint)(lba % logicalSectors);            
            pb.IsLogical = true;

            //Console.WriteLine("{0}", pb);

            return pb;
        } 

        public override Block LogicalBlockToPhysicalBlock(Block logBlock)
        {

            Block physBlock = new Block();

            // This is really quite odd.
            // The gist of this seems to be (and the code below is not quite correct, logically)
            // that on Track 1, logical cylinders 0-71 are physically translated 5 cylinders in;
            // but the last 5 cylinders (logical 72-76) are mapped to cylinders 0-4.  Weird.
            if (logBlock.Cylinder < 72)
            {
                // Move this forward 5 tracks to compensate for the logical->physical conversion
                physBlock.Cylinder = logBlock.Cylinder + 5;
            }
            else
            {
                physBlock.Cylinder = logBlock.Cylinder % 72;
            }

            // there are four physical blocks per logical block; block 0 starts at sector 3.
            physBlock.Sector = ((logBlock.Sector) * 4) + 2;

            physBlock.Track = logBlock.Track;

            if (physBlock.Cylinder >= Cylinders)
            {
                throw new InvalidOperationException("Went off the end of the disk converting to physical.");
            }

            if (physBlock.Sector >= Sectors)
            {
                // this should never happen in normal operation
                throw new InvalidOperationException("Went off the end of the track converting to physical.");
            }

            physBlock.IsLogical = false;

            return physBlock;
        }
      

        /// <summary>
        /// HACKY:
        /// Returns a 256-word block (1 logical block = 4 physical sectors) starting at the given sector...
        /// This code assumes that 4 sectors are available starting at the given sector,
        /// otherwise bad things will happen.
        /// </summary>
        /// <param name="cylinder"></param>
        /// <param name="track"></param>
        /// <param name="sector"></param>
        /// <returns></returns>
        private Sector CreateLogicalBlock(uint cylinder, uint track, uint sector)
        {
            uint logicalSector = (sector - 2) / 4;

            Sector logical = new Sector((uint)cylinder, (uint)track, (uint)logicalSector, new DiskGeometry(0, 0, 0, 512, false));

            for (int i = 0; i < 4; i++)
            {
                Sector physical =
                    _sectors[cylinder,
                             track,
                             _sectorOrder[logicalSector, i] - 1];

                physical.Data.CopyTo(logical.Data, i * _diskType.SectorSize);

            }


            return logical;
        }

        private BlockHeader[,,] _header;

        /// <summary>
        /// The orderings for physical sectors for each logical sector.
        /// These are not in linear order (they appear to be interleaved for speed optimization) 
        /// and the order isn't actually documented anywhere that I can find...
        /// Thse numbers were deduced by seeing how the POS code accesses the disk through PERQemu...
        /// </summary>
        private int[,] _sectorOrder = {
            {3, 8, 13, 18},
            {23, 4, 9, 14},
            {19, 24, 5, 10},
            {15, 20, 25, 6},
            {11, 16, 21, 26},
            {7, 12, 17, 22} };
    }
}
