using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace PERQDisk
{
    /// <summary>
    /// Represents the physical structure of a given disk (shugart/micropolis/floppy, etc...)
    /// </summary>
    public struct DiskGeometry
    {
        public DiskGeometry(uint cyls, uint tracks, uint secs, uint sectorSize, bool hasBlockHeader)
        {
            if (cyls < 0 || tracks < 0 || secs < 0)
            {
                throw new ArgumentOutOfRangeException("Invalid chs specification, must be greater than zero.");
            }

            _cylinders = cyls;
            _tracks = tracks;
            _sectors = secs;
            _sectorSize = sectorSize;
            _hasBlockHeader = hasBlockHeader;
        }

        public uint Cylinders
        {
            get { return _cylinders; }
        }

        public uint Tracks
        {
            get { return _tracks; }
        }

        public uint Sectors
        {
            get { return _sectors; }
        }

        public uint SectorSize
        {
            get { return _sectorSize; }
        }

        public bool HasBlockHeader
        {
            get { return _hasBlockHeader; }
        }

        private uint _cylinders;
        private uint _tracks;
        private uint _sectors;
        private uint _sectorSize;
        private bool _hasBlockHeader;

        public static DiskGeometry Shugart12 = new DiskGeometry(202, 4, 30, 512, true);
        public static DiskGeometry Shugart24 = new DiskGeometry(202, 8, 30, 512, true);
        public static DiskGeometry Micropolis = new DiskGeometry(580, 5, 24, 512, true);
        public static DiskGeometry FloppySS = new DiskGeometry(77, 1, 26, 128, false);
        public static DiskGeometry FloppyDS = new DiskGeometry(77, 2, 26, 128, false);
    }

    /// <summary>
    /// Represents a cyl,track,sector triad; additionally contains a bit indicating whether this
    /// block is "logical" (excluding boot sector) or "physical" (representing actual disk layout).
    /// This could (should) probably be factored into two classes, inheriting from a common abstract base class.
    /// But I'm lazy.
    /// </summary>
    public struct Block
    {
        public Block(uint cylinder, uint track, uint sector, bool isLogical)
        {
            Cylinder = cylinder;
            Track = track;
            Sector = sector;
            IsLogical = isLogical;
        }

        public uint Cylinder;
        public uint Track;
        public uint Sector;
        public bool IsLogical;

        public override string ToString()
        {
            if (IsLogical)
            {
                return String.Format("Logical Cylinder {0}, Track {1}, Sector {2}", Cylinder, Track, Sector);
            }
            else
            {
                return String.Format("Physical Cylinder {0}, Track {1}, Sector {2}", Cylinder, Track, Sector);
            }
        }
    }

    /// <summary>
    /// Represents a physical disk drive, which contains one or more Cylinders.
    /// This is the primary interface for interaction with the disk image, and allows
    /// reading sectors of data.    
    /// </summary>
    public abstract class PhysicalDisk
    {
        /// <summary>
        /// Constructs a new PhysicalDisk, given the number of Cylinders on the drive,
        /// Tracks per Cylinder and Sectors per Track.
        /// </summary>
        /// <param name="cylinders"></param>
        /// <param name="tracks"></param>
        /// <param name="sectors"></param>
        public PhysicalDisk()
        {            
            _loaded = false;
        }


        public bool Loaded
        {
            get { return _loaded; }
        }

        /// <summary>
        /// Loads the disk image from the supplied filestream.
        /// </summary>
        /// <param name="fs"></param>
        public virtual void Load(FileStream fs)
        {
            for (int cyl = 0; cyl < _diskType.Cylinders; cyl++)
            {
                for (int track = 0; track < _diskType.Tracks; track++)
                {
                    for (int sector = 0; sector < _diskType.Sectors; sector++)
                    {
                        _sectors[cyl, track, sector].Load(fs);
                    }
                }
            }

            _loaded = true;
        }

        public virtual Sector GetSector(uint cylinder, uint track, uint sector)
        {
            if (cylinder < 0 || cylinder >= _diskType.Cylinders)
            {
                throw new ArgumentOutOfRangeException("cylinder");
            }

            if (track < 0 || track >= _diskType.Tracks)
            {
                throw new ArgumentOutOfRangeException("track");
            }

            if (sector < 0 || sector >= _diskType.Sectors)
            {
                throw new ArgumentOutOfRangeException("sector");
            }

            return _sectors[cylinder, track, sector];            
        }

        public virtual Sector GetSector(Block block)
        {
            // Get physical block if this is a logical one.
            if (block.IsLogical)
            {
                block = LogicalBlockToPhysicalBlock(block);
            }
            
            return GetSector(block.Cylinder, block.Track, block.Sector);            
        }

        /// <summary>
        /// Converts a Logical Disk Address (Block #) to a logical block (cyl/track/sector) specification.
        /// This is drive-specific.
        /// </summary>
        /// <param name="lda"></param>
        /// <returns></returns>
        public abstract Block LDAToLogicalBlock(uint lda);               

        /// <summary>
        /// Converts a Logical block mapping to a Physical block mapping.
        /// This is drive-specific.
        /// </summary>
        /// <param name="physBlock"></param>
        /// <returns></returns>
        public abstract Block LogicalBlockToPhysicalBlock(Block physBlock);

        /// <summary>
        /// Indicates the geometry of the drive
        /// </summary>
        public DiskGeometry DiskGeometry
        {
            get { return _diskType; }
        }

        public uint Cylinders
        {
            get { return _diskType.Cylinders; }
        }

        public uint Tracks
        {
            get { return _diskType.Tracks; }
        }

        public uint Sectors
        {
            get { return _diskType.Sectors; }
        }

        protected void CreateSectors()
        {
            _sectors = new Sector[Cylinders, Tracks, Sectors];
            

            for (uint cyl = 0; cyl < Cylinders; cyl++)
            {
                for (uint track = 0; track < Tracks; track++)
                {
                    for (uint sector = 0; sector < Sectors; sector++)
                    {
                        _sectors[cyl, track, sector] = new Sector(cyl, track, sector, DiskGeometry);
                    }
                }
            }
        }

        protected void VerifyPhysical(Block physBlock)
        {
            if (physBlock.IsLogical)
            {
                throw new InvalidOperationException("This block is not physical.");
            }
        }

        protected void VerifyLogical(Block logBlock)
        {
            if (!logBlock.IsLogical)
            {
                throw new InvalidOperationException("This block is not logical.");
            }
        }

        protected Sector[, ,] _sectors;
        
        protected DiskGeometry _diskType;

        private bool _loaded;
    }

}
