using System;
using System.Collections.Generic;
using System.Text;

namespace PERQDisk
{
    // Represents a 24mb Shugart hard disk.
    public class ShugartDisk : PhysicalDisk
    {
        public ShugartDisk(bool is24mb) : base()
        {
            _diskType = is24mb ? DiskGeometry.Shugart24 : DiskGeometry.Shugart12;
            
            CreateSectors();
        }

        /// <summary>
        /// Converts a logical disk address into a Logical block expressed in Cyl,Track,Sector format.
        /// 
        /// This routine converts a LDA (logical disk address) to an LBN; this can then be directly converted
        /// to a Physical block address (LBN's cyl0,track0,sector0 is physical cyl0,track1,sector0)
        /// Summarized from the POS documentation, Page 1-7:       
        /// 
        /// Low word:
        ///     - Bits 0 through 7 are reserved for future use
        ///     - Bits 8 through 15 form the low 8 bits of the LBN
        /// 
        /// High word:
        ///     - Bits 0 through 10 form the high eleven bits of the LBN
        ///     - Bits 11 through 13 specify the volume (000 = hard disk, 100 = floppy)
        ///     - Bits 14 and 15 indicate whether the address is on disk or virtualized.
        /// 
        /// Currently we assume a hard disk image and so only the LBN-specific data is taken into account here.
        ///         
        /// 
        /// </summary>
        /// <param name="logicalAddress"></param>
        /// <returns></returns>
        public override Block LDAToLogicalBlock(uint lda)
        {
            uint lba = ((lda & 0x0000ff00) >> 8) | ((lda & 0x07ff0000) >> 8);

            //Now we have the actual logical block number. 

            Block pb = new Block();
            pb.Cylinder = (uint)(lba / (Tracks * Sectors));
            pb.Track = (uint)((lba - (pb.Cylinder * Tracks * Sectors)) / Sectors);
            pb.Sector = (uint)(lba % Sectors);
            pb.IsLogical = true;

            return pb;
        }


        public override Block LogicalBlockToPhysicalBlock(Block logBlock)
        {

            VerifyLogical(logBlock);

            // Move this forward one track to compensate for the logical->physical conversion
            logBlock.Track = logBlock.Track + 1;

            if (logBlock.Track >= Tracks)
            {
                logBlock.Track = 0;
                logBlock.Cylinder = logBlock.Cylinder + 1;

                if (logBlock.Cylinder >= Cylinders)
                {
                    throw new InvalidOperationException("Went off the end of the disk converting to physical.");
                }
            }

            logBlock.IsLogical = false;

            return logBlock;
        }       
    }
}
